"""Information about the contents of a Task output archive."""
from dataclasses import dataclass
from typing import List

from inductiva.utils import format_utils


@dataclass
class FileInfo:
    """Information about a file in an output archive."""
    name: str
    size: int
    compressed_size: int


@dataclass
class OutputContents:
    """Information about the contents of a Task output archive."""
    size: int
    contents: List[FileInfo]

    @property
    def num_files(self) -> int:
        return len(self.contents)

    def __str__(self) -> str:
        size_header = "Size"
        compressed_header = "Compressed"
        name_header = "Name"

        lines = [
            f"Archive size: {format_utils.bytes_formatter(self.size)}",
            "Contents:",
            f"  {size_header:<12} {compressed_header:<12} {name_header:<12}",
        ]
        for file in self.contents:
            lines.append(
                f"  {format_utils.bytes_formatter(file.size):<12}"
                f" {format_utils.bytes_formatter(file.compressed_size):<12}"
                f" {file.name:<12}")

        return "\n".join(lines)



"""XBeach module of the API."""
from typing import Optional

from inductiva import types, tasks, simulators


@simulators.simulator.mpi_enabled
class XBeach(simulators.Simulator):
    """Class to invoke a generic XBeach simulation on the API."""

    def __init__(self):
        super().__init__()
        self.api_method_name = "sw.xbeach.run_simulation"

    def run(self,
            input_dir: types.Path,
            n_vcpus: Optional[int] = None,
            use_hwthread: bool = True,
            sim_config_filename: Optional[str] = "params.txt",
            on: Optional[types.ComputationalResources] = None,
            storage_dir: Optional[types.Path] = "",
            extra_metadata: Optional[dict] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            sim_config_filename: Name of the simulation configuration file.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
            (default), all vCPUs will be used.
            use_hwthread: If specified Open MPI will attempt to discover the
            number of hardware threads on the node, and use that as the
            number of slots available.
            on: The computational resource to launch the simulation on. If None
                the simulation is submitted to a machine in the default pool.
            storage_dir: Directory for storing simulation results.
            other arguments: See the documentation of the base class.
        """
        return super().run(input_dir,
                           input_filename=sim_config_filename,
                           on=on,
                           storage_dir=storage_dir,
                           n_vcpus=n_vcpus,
                           use_hwthread=use_hwthread,
                           extra_metadata=extra_metadata)



# pylint: disable=missing-module-docstring
from .simulator import Simulator
from .swash import SWASH
from .xbeach import XBeach
from .splishsplash import SplishSplash
from .dualsphysics import DualSPHysics
from .openfoam import OpenFOAM
from .gromacs import GROMACS
from .simsopt import SIMSOPT
from .fenicsx import FEniCSx
from .fds import FDS
from .reef3d import REEF3D
from .dummy_simulator import DummySimulator
from .swan import SWAN
from .schism import SCHISM



"""Methods that interact with the lower-level inductiva-web-api-client.

The relevant function for usage outside of this file is invoke_async_api().
Check the demos directory for examples on how it is used.
"""
import os
import pathlib
import signal
import time
from urllib3.exceptions import MaxRetryError, NewConnectionError
from contextlib import contextmanager
from typing import Any, Dict, List, Optional, Tuple, Type

from absl import logging

import inductiva
from inductiva.client import ApiClient, ApiException, Configuration
from inductiva.client.apis.tags.tasks_api import TasksApi
from inductiva.client.apis.tags.version_api import VersionApi
from inductiva.client.models import (BodyUploadTaskInput, TaskRequest,
                                     TaskStatus)
from inductiva import types, constants
from inductiva.utils.data import (extract_output, get_validate_request_params,
                                  pack_input)
from inductiva.utils import format_utils, files


def validate_api_key(api_key: Optional[str]) -> Configuration:
    """Validates the API key and returns API configuration"""
    if inductiva.api_key is None:
        # pylint: disable=line-too-long
        raise ValueError(
            "No API Key specified. "
            "Please set the INDUCTIVA_API_KEY environment variable.\n"
            "More infomation at:"
            "https://docs.inductiva.ai/")
    # pylint: enable=line-too-long

    # Perform version check only on first invocation
    if not hasattr(validate_api_key, "version_checked"):
        compare_client_and_backend_versions(inductiva.__version__)
        validate_api_key.version_checked = True

    api_config = Configuration(host=inductiva.api_url)
    api_config.api_key["APIKeyHeader"] = api_key

    return api_config


def get_client() -> ApiClient:
    """Returns an ApiClient instance."""
    api_config = validate_api_key(inductiva.api_key)

    return ApiClient(api_config)


def submit_request(api_instance: TasksApi, request: TaskRequest) -> TaskStatus:
    """Submits a task request to the API.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
    Return:
        Returns the body of the HTTP response.
        Contains two fields, "id" and "status".
    """

    try:
        api_response = api_instance.submit_task(body=request)
    except ApiException as e:
        logging.exception("Exception when calling TasksApi->submit_task: %s", e)
        raise e

    logging.debug("Request status: %s", api_response.body["status"])

    return api_response.body


def upload_input(api_instance: TasksApi, task_id, original_params,
                 type_annotations):
    """Uploads the inputs of a given task to the API.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task.
        original_params: Params of the request passed by the user.
        type_annotations: Annotations of the params' types.
    """

    inputs_size = files.get_path_size(original_params["sim_dir"])
    logging.info("Preparing upload of the local input directory %s (%s).",
                 original_params["sim_dir"],
                 format_utils.bytes_formatter(inputs_size))
    input_zip_path = pack_input(
        params=original_params,
        type_annotations=type_annotations,
        zip_name=task_id,
    )

    try:
        with open(input_zip_path, "rb") as zip_fp:
            _ = api_instance.upload_task_input(
                path_params={"task_id": task_id},
                body=BodyUploadTaskInput(file=zip_fp),
            )
    except ApiException as e:
        logging.exception(
            "Exception when calling TasksApi->upload_task_inputs: %s", e)
        raise e

    logging.info("Local input directory successfully uploaded.")

    os.remove(input_zip_path)


def download_output(
        api_instance: TasksApi,
        task_id,
        output_dir: Optional[types.Path] = None) -> Tuple[List, pathlib.Path]:
    """Downloads the output of a given task from the API.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task.

    Return:
        Downloads and extracts the data to the client.
    """

    if output_dir is None:
        output_dir = os.path.join(inductiva.get_output_dir(), task_id)

    logging.info("Downloading the task %s outputs to %s...", task_id,
                 output_dir)
    try:
        api_response = api_instance.download_task_output(
            path_params={"task_id": task_id},
            stream=True,
        )
    except ApiException as e:
        raise e

    logging.debug("Downloaded output to %s", api_response.body.name)

    result_list = extract_output(api_response.body.name, output_dir)
    logging.info("Task %s output successfully downloaded to %s.", task_id,
                 output_dir)

    return result_list, pathlib.Path(output_dir)


def block_until_finish(api_instance: TasksApi, task_id: str) -> str:
    """Block until a task executing remotely finishes execution.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task to wait for.

    Returns:
        Returns info related to the task, containing two fields,
        "id" and "status".
    """
    logging.debug("Blocking until task is finished ...")
    return block_until_status_is(api_instance, task_id, {"success", "failed"})


def kill_task(api_instance: TasksApi, task_id: str):
    """Kill a task that is executing remotely.

    The function sends a kill request to the API.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task to kill.
   """
    logging.debug("Sending kill task request ...")
    api_instance.kill_task(path_params={"task_id": task_id},)
    logging.info("Task terminated.")


def get_task_status(api_instance: TasksApi, task_id: str) -> TaskStatus:
    """Check the status of a task."""

    api_response = api_instance.get_task_status(
        path_params={"task_id": task_id})

    status = api_response.body["status"]

    return status


def block_until_status_is(api_instance: TasksApi,
                          task_id,
                          desired_status,
                          sleep_secs=0.5):
    """Block until the status of a task becomes the desired status.

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task to wait for.
        desired_status: Task status to wait for.
        sleep_secs: Polling interval.

    Returns:
        Returns info related to the task, containing two fields,
    """
    prev_status = None

    while True:
        try:
            status = get_task_status(api_instance, task_id)
            logging.debug("Task status is %s", status)
        except ApiException as e:
            raise e

        if status != prev_status:
            if status == "submitted":
                logging.info("Waiting for resources...")

            prev_status = status

        # If status reaches the desired status, then stop polling
        if status in desired_status:
            break

        time.sleep(sleep_secs)

    return status


@contextmanager
def blocking_task_context(api_instance: TasksApi, task_id):
    """Context to handle execution of a blocking task.

    The context handles exceptions and the SIGINT signal, issuing a request
    to the API to kill the executing task.
    Info on the implementation:
    - https://docs.python.org/3/library/contextlib.html

    Args:
        api_instance: Instance of TasksApi used to send necessary requests.
        task_id: ID of the task being executed.
    """
    # Other imported modules can make changes to the SIGINT handler, so we set
    # the default int handler to make sure that KeyboardInterrupt is raised
    # if the user presses Ctrl+C.
    original_sig = signal.signal(signal.SIGINT, signal.default_int_handler)

    try:
        yield None
    except Exception as err:
        logging.info("Caught exception: terminating blocking task...")
        kill_task(api_instance, task_id)
        raise err
    except KeyboardInterrupt as err:
        logging.info("Caught SIGINT: terminating blocking task...")
        kill_task(api_instance, task_id)
        raise err
    finally:
        # Reset original SIGINT handler
        signal.signal(signal.SIGINT, original_sig)


def log_task_info(task_id, method_name, params, resource_pool):
    """Logging the main components of a task submission."""

    logging.info("Task Information:")
    logging.info("> ID:                    %s", task_id)
    logging.info("> Method:                %s", method_name.split(".")[1])
    logging.info("> Local input directory: %s", params["sim_dir"])
    logging.info("> Submitting to the following computational resources:")
    if resource_pool is not None:
        logging.info(" >> %s", resource_pool)
    else:
        logging.info(" >> Default queue with %s machines.",
                     constants.DEFAULT_QUEUE_MACHINE_TYPE)


def submit_task(api_instance, method_name, request_params, resource_pool,
                storage_path_prefix, params, type_annotations):
    """Submit a task and send input files to the API."""

    resource_pool_id = None
    if resource_pool is not None:
        resource_pool_id = resource_pool.id

    task_request = TaskRequest(method=method_name,
                               params=request_params,
                               resource_pool=resource_pool_id,
                               storage_path_prefix=storage_path_prefix)
    task = submit_request(
        api_instance=api_instance,
        request=task_request,
    )

    task_id = task["id"]
    log_task_info(task_id, method_name, params, resource_pool)

    if task["status"] == "pending-input":

        upload_input(
            api_instance=api_instance,
            original_params=params,
            task_id=task_id,
            type_annotations=type_annotations,
        )

    return task_id


def invoke_async_api(method_name: str,
                     params,
                     type_annotations: Dict[Any, Type],
                     resource_pool: Optional[
                         types.ComputationalResources] = None,
                     storage_path_prefix: Optional[str] = "") -> str:
    """Perform a task asyc and remotely via Inductiva's Web API.

    Submits a simulation async to the API and returns the task id.
    The flow is summarized as follows:
        1. Transform request params into the params used to
        validate permission to execute the request.
        2. Submit task via the "POST task/submit" endpoint.
            Note: HTTP status code 400 informs the requested method is invalid,
                and HTTP status code 403 informs that the user is not authorized
                to post such request.
        3. If the status returned by the previous HTTP request is
            "pending-input", ZIP inputs and send them via
            "POST task/{task_id}/input".
        4. Return task_id and leaves the simulation on the queue until resources
            become available.

    Args:
        request: Request sent to the API for validation.
        input_dir: Directory containing the input files to be uploaded.

    Return:
        Returns the task id.
    """

    api_config = validate_api_key(inductiva.api_key)

    request_params = get_validate_request_params(
        original_params=params,
        type_annotations=type_annotations,
    )

    with ApiClient(api_config) as client:
        api_instance = TasksApi(client)

        task_id = submit_task(api_instance=api_instance,
                              method_name=method_name,
                              request_params=request_params,
                              resource_pool=resource_pool,
                              storage_path_prefix=storage_path_prefix,
                              params=params,
                              type_annotations=type_annotations)

    return task_id


def compare_client_and_backend_versions(client_version: str):
    """ Compares the provided client version 7with the backend API version.

    Sends a GET request to the backend API's version comparison endpoint
    with the client version as a parameter. Evaluates the response to
    determine if the client version is compatible with the backend version.
    Raises exceptions for communication issues or incompatibility.

    Parameters:
    - client_version (str): The version of the client to be compared with the
                            backend version.

    Raises:
    - RuntimeError: If the API cannot be reached, or if the client version is
      incompatible with the backend version, or for other general failures.
    """
    api_config = Configuration(host=inductiva.api_url)

    with ApiClient(api_config) as client:
        api_instance = VersionApi(client)
        query_params = {"client_version": client_version}

        try:
            api_instance.compare_client_and_backend_versions(
                query_params=query_params)

        except (MaxRetryError, NewConnectionError) as exc:
            raise RuntimeError(
                "Failed to reach the API. "
                "Please check your connection and try again.") from exc

        except ApiException as e:
            if e.status == 406:
                raise RuntimeError(
                    f"Client version {client_version} is not compatible "
                    f"with API version {e.headers['version']}.\n"
                    "Please update the client version.") from e
            raise RuntimeError(e) from e

        except Exception as e:
            raise RuntimeError(
                f"Failed to compare client and API versions. {e}") from e



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class MachineType(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "machine_type",
            "num_cpus",
            "price",
            "spot",
            "region",
            "ram_gb",
        }

        class properties:
            machine_type = schemas.StrSchema
            num_cpus = schemas.IntSchema
            ram_gb = schemas.IntSchema
            price = schemas.NumberSchema
            spot = schemas.BoolSchema
            region = schemas.StrSchema

            class provider(schemas.EnumBase, schemas.StrSchema):

                class MetaOapg:
                    enum_value_to_name = {
                        "GCP": "GCP",
                        "ICE": "ICE",
                    }

                @schemas.classproperty
                def GCP(cls):
                    return cls("GCP")

                @schemas.classproperty
                def ICE(cls):
                    return cls("ICE")

            __annotations__ = {
                "machine_type": machine_type,
                "num_cpus": num_cpus,
                "ram_gb": ram_gb,
                "price": price,
                "spot": spot,
                "region": region,
                "provider": provider,
            }

    machine_type: MetaOapg.properties.machine_type
    num_cpus: MetaOapg.properties.num_cpus
    price: MetaOapg.properties.price
    spot: MetaOapg.properties.spot
    region: MetaOapg.properties.region
    ram_gb: MetaOapg.properties.ram_gb

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["machine_type"]
    ) -> MetaOapg.properties.machine_type:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["num_cpus"]
    ) -> MetaOapg.properties.num_cpus:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["ram_gb"]
    ) -> MetaOapg.properties.ram_gb:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["price"]
    ) -> MetaOapg.properties.price:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["spot"]
    ) -> MetaOapg.properties.spot:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["region"]
    ) -> MetaOapg.properties.region:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["provider"]
    ) -> MetaOapg.properties.provider:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "machine_type",
        "num_cpus",
        "ram_gb",
        "price",
        "spot",
        "region",
        "provider",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["machine_type"]
    ) -> MetaOapg.properties.machine_type:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["num_cpus"]
    ) -> MetaOapg.properties.num_cpus:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["ram_gb"]
    ) -> MetaOapg.properties.ram_gb:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["price"]
    ) -> MetaOapg.properties.price:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["spot"]
    ) -> MetaOapg.properties.spot:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["region"]
    ) -> MetaOapg.properties.region:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["provider"]
    ) -> typing.Union[MetaOapg.properties.provider, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "machine_type",
        "num_cpus",
        "ram_gb",
        "price",
        "spot",
        "region",
        "provider",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        machine_type: typing.Union[
            MetaOapg.properties.machine_type,
            str,
        ],
        num_cpus: typing.Union[
            MetaOapg.properties.num_cpus,
            decimal.Decimal,
            int,
        ],
        price: typing.Union[
            MetaOapg.properties.price,
            decimal.Decimal,
            int,
            float,
        ],
        spot: typing.Union[
            MetaOapg.properties.spot,
            bool,
        ],
        region: typing.Union[
            MetaOapg.properties.region,
            str,
        ],
        ram_gb: typing.Union[
            MetaOapg.properties.ram_gb,
            decimal.Decimal,
            int,
        ],
        provider: typing.Union[MetaOapg.properties.provider, str,
                               schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'MachineType':
        return super().__new__(
            cls,
            *_args,
            machine_type=machine_type,
            num_cpus=num_cpus,
            price=price,
            spot=spot,
            region=region,
            ram_gb=ram_gb,
            provider=provider,
            _configuration=_configuration,
            **kwargs,
        )



"""OpenFOAM module of the API for fluid dynamics."""
from typing import Optional

from inductiva import types, tasks, simulators

AVAILABLE_OPENFOAM_VERSIONS = ["foundation", "esi"]


@simulators.simulator.mpi_enabled
class OpenFOAM(simulators.Simulator):
    """Class to invoke a generic OpenFOAM simulation on the API.

    Users can choose between the ESI or the Foundation version
    by selecting the version on the initiliasation. Be aware, that
    some input files may only work for a specific version.
    """

    def __init__(self, version: str = "foundation"):
        if version not in AVAILABLE_OPENFOAM_VERSIONS:
            raise ValueError("Version not currently supported."
                             f"Available: {AVAILABLE_OPENFOAM_VERSIONS}")

        super().__init__()
        self.api_method_name = f"fvm.openfoam_{version}.run_simulation"

    def run(self,
            input_dir: types.Path,
            commands: types.Commands,
            n_vcpus: Optional[int] = None,
            use_hwthread: bool = True,
            on: Optional[types.ComputationalResources] = None,
            storage_dir: Optional[types.Path] = "",
            extra_metadata: Optional[dict] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            commands: List of commands to run using the OpenFOAM simulator.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
            (default), all vCPUs will be used.
            use_hwthread: If specified Open MPI will attempt to discover the
            number of hardware threads on the node, and use that as the
            number of slots available.
            on: The computational resource to launch the simulation on. If None
                the simulation is submitted to a machine in the default pool.
            other arguments: See the documentation of the base class.
        """
        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           n_vcpus=n_vcpus,
                           use_hwthread=use_hwthread,
                           extra_metadata=extra_metadata)



"""Base class for machine groups."""
import time
import enum
from typing import Union
from abc import abstractmethod
import datetime

import logging

import inductiva
import inductiva.client.models
from inductiva import api
from inductiva.utils import format_utils
from inductiva.client.apis.tags import compute_api
from inductiva.client import exceptions
from inductiva import logs

from inductiva.resources import machine_types


class ResourceType(enum.Enum):
    """Enum to represent the type of machine to be launched."""

    STANDARD = "standard"
    MPI = "mpi"


class BaseMachineGroup:
    """Base class to manage Google Cloud resources."""

    def __init__(self,
                 machine_type: str,
                 provider: Union[machine_types.ProviderType, str] = "GCP",
                 data_disk_gb: int = 10,
                 register: bool = True) -> None:
        """Create a BaseMachineGroup object.

        Args:
            machine_type: The type of GC machine to launch. Ex: "e2-standard-4".
              Check https://cloud.google.com/compute/docs/machine-resource for
              more information about machine types.
            data_disk_gb: The size of the disk for user data (in GB).
            register: Bool that indicates if a machine group should be register
                or if it was already registered. If set to False by users on
                initialization, then, the machine group will not be able to be
                started. This serves has an helper argument for retrieving
                already registered machine groups that can be started, for
                example, when retrieving with the `machines_groups.get` method.
                Users should not set this argument in anyway.
        """

        provider = machine_types.ProviderType(provider)
        self.provider = provider.value

        if machine_type not in machine_types.list_available_machines(
                self.provider):
            raise ValueError(f"Machine type not supported in {self.provider}")

        if data_disk_gb <= 0:
            raise ValueError("`data_disk_gb` must be positive.")

        self.machine_type = machine_type
        self.provider = provider.value
        self.data_disk_gb = data_disk_gb
        self._id = None
        self._name = None
        self.create_time = None
        self._started = False
        self.register = register
        #Number of active machines at the time of
        #the request machine_groups.get()
        self._active_machines = 0
        self.num_machines = 0

        # Set the API configuration that carries the information from the client
        # to the backend.
        self._api = compute_api.ComputeApi(api.get_client())
        self._estimated_cost = None

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    def _register_machine_group(self, **kwargs):
        """Register machine group configuration in API.
        
        Returns:
            The unique ID and name identifying the machine on the API."""

        instance_group_config = inductiva.client.models.VMGroupConfig(
            machine_type=self.machine_type,
            provider_id=self.provider,
            disk_size_gb=self.data_disk_gb,
            **kwargs,
        )

        try:
            resp = self._api.register_vm_group(body=instance_group_config)
        except (exceptions.ApiValueError, exceptions.ApiException) as e:
            logs.log_and_exit(
                logging.getLogger(),
                logging.ERROR,
                "Registering machine group failed with exception %s",
                e,
                exc_info=e)

        self._id = resp.body["id"]
        self._name = resp.body["name"]
        self.register = False
        self._log_machine_group_info()

    @abstractmethod
    def __repr__(self):
        pass

    def active_machines_to_str(self) -> str:
        """Return the number of machines currently running.
        """
        return f"{self._active_machines}/{self.num_machines}"

    @classmethod
    def from_api_response(cls, resp: dict):
        """Creates a MachineGroup object from an API response."""

        machine_group = cls(
            machine_type=resp["machine_type"],
            data_disk_gb=resp["disk_size_gb"],
            register=False,
        )
        machine_group._id = resp["id"]
        machine_group.provider = resp["provider_id"]
        machine_group._name = resp["name"]
        machine_group.create_time = resp["creation_timestamp"]
        machine_group._started = True

        return machine_group

    def update_termination_timers(self,
                                  max_idle_time: datetime.timedelta = None,
                                  auto_terminate_ts: datetime.datetime = None):
        """Update the termination timers of a machine group.

        Args:
            max_idle_time (timedelta): Max idle time, i.e. time without
                executing any task, after which the resource will be terminated.
            auto_terminate_ts (datetime): Moment in which the resource will
                be automatically terminated, irrespectively of the existence of
                tasks yet to be executed by the resource.
        """

        # Convert max_idle_time from minutes to seconds
        max_idle_time = max_idle_time.total_seconds() if max_idle_time else None

        # Convert auto_terminate_ts to ISO format
        if auto_terminate_ts is not None:
            now_ts = datetime.datetime.now(auto_terminate_ts.tzinfo)
            if auto_terminate_ts < now_ts:
                raise ValueError("auto_terminate_ts must be in the future.")
            auto_terminate_ts = auto_terminate_ts.isoformat()

        update_body = inductiva.client.models.VMGroupLifecycleConfig(
            max_idle_time=max_idle_time, auto_terminate_ts=auto_terminate_ts)

        if max_idle_time is not None or auto_terminate_ts is not None:
            try:
                self._api.update_vm_group_config(
                    path_params={"mg_id": self._id}, body=update_body)
            except inductiva.client.ApiException as e:
                logs.log_and_exit(
                    logging.getLogger(),
                    logging.ERROR,
                    "Setting termination timers for machine group failed " \
                    "with exception %s.",
                    e,
                    exc_info=e)

    def start(self,
              max_idle_time: datetime.timedelta = None,
              auto_terminate_ts: datetime.datetime = None,
              **kwargs):
        """Starts a machine group.

        Args:
            max_idle_time (timedelta): Max idle time, i.e. time without
                executing any task, after which the resource will be terminated.
            auto_terminate_ts (datetime): Moment in which the resource will
                be automatically terminated, irrespectively of the existence of
                tasks yet to be executed by the resource.
            **kwargs: Depending on the type of machine group to be started,
              this can be num_machines, max_machines, min_machines,
              and is_elastic."""
        if self._started:
            logging.info("Attempting to start a machine group already started.")
            return

        if self.id is None or self.name is None:
            logging.info("Attempting to start an unregistered machine group. "
                         "Make sure you have called the constructor without "
                         "`register=False`.")
            return

        request_body = \
            inductiva.client.models.VMGroupConfig(
                id=self.id,
                name=self.name,
                machine_type=self.machine_type,
                provider_id=self.provider,
                disk_size_gb=self.data_disk_gb,
                **kwargs,
            )

        logging.info("Starting %s. "
                     "This may take a few minutes.", repr(self))
        logging.info("Note that stopping this local process will not interrupt "
                     "the creation of the machine group. Please wait...")
        start_time = time.time()

        self.update_termination_timers(max_idle_time, auto_terminate_ts)

        try:

            self._api.start_vm_group(body=request_body)
        except inductiva.client.ApiException as e:
            logs.log_and_exit(logging.getLogger(),
                              logging.ERROR,
                              "Starting machine group failed with exception %s",
                              e,
                              exc_info=e)
        creation_time = format_utils.seconds_formatter(time.time() - start_time)
        self._started = True
        logging.info("%s successfully started in %s.", self, creation_time)

    def terminate(self, **kwargs):
        """Terminates a machine group."""
        if not self._started or self.id is None or self.name is None:
            logging.info("Attempting to terminate an unstarted machine group.")
            return

        try:
            logging.info("Terminating %s. This may take a few minutes.",
                         repr(self))
            start_time = time.time()

            request_body = \
                inductiva.client.models.VMGroupConfig(
                    id=self.id,
                    name=self.name,
                    machine_type=self.machine_type,
                    provider_id=self.provider,
                    disk_size_gb=self.data_disk_gb,
                    **kwargs,
                )

            self._api.delete_vm_group(body=request_body)
            termination_time = format_utils.seconds_formatter(time.time() -
                                                              start_time)
            logging.info("%s successfully terminated in %s.", self,
                         termination_time)

        except inductiva.client.ApiException as api_exception:
            raise api_exception

    def _get_estimated_cost(self, spot: bool = False) -> float:
        """Returns estimate cost of a single machine in the group.

        This method is an overlay of the more general method, but
        it verifies if the cost has already been estimated and returns
        it immediately if it has.
        """
        if self.provider == "ICE":
            return 0

        if self._estimated_cost is not None:
            return self._estimated_cost

        self._estimated_cost = inductiva.resources.estimate_machine_cost(
            self.machine_type,
            spot,
        )

        return self._estimated_cost

    def status(self):
        """Returns the status of a machine group if it exists.

        Otherwise returns None"""
        if self.name is None:
            logging.info(
                "Attempting to get the status of an unregistered machine group."
            )
            return

        response = self._api.get_group_status({"name": self.name})
        if response.body == "notFound":
            logging.info("Machine group does not exist: %s.", self.name)
        return response.body

    def _log_machine_group_info(self):
        """Logs the machine group info."""

        logging.info("> Name:         %s", self.name)
        logging.info("> Machine Type: %s", self.machine_type)
        logging.info("> Data disk size:    %s GB", self.data_disk_gb)



"""Base and specific renderers packages to manage template files."""
from typing import Union, IO
from abc import abstractmethod
import pathlib
import os

import jinja2

from inductiva import types


class RendererAdapter:
    """Abstract base class for template renderers."""

    TEMPLATE_EXTENSION = None

    @abstractmethod
    def set_template_dir(self, template_dir: types.PathOrStr) -> None:
        """Set the directory where the templates are located."""
        pass

    @abstractmethod
    def render(self, template_name: types.PathOrStr,
               target_file: Union[types.PathOrStr, IO], **render_args):
        """Render a file."""
        pass

    @classmethod
    def is_template(cls, file: types.PathOrStr) -> bool:
        """Check if the given file is of template type."""
        if isinstance(file, pathlib.Path):
            return file.suffix == cls.TEMPLATE_EXTENSION
        elif isinstance(file, str):
            return file.endswith(cls.TEMPLATE_EXTENSION)
        return False

    @classmethod
    def strip_extension(cls, file: types.PathOrStr) -> str:
        """Strip the template extension if the given file is a template."""
        if not cls.is_template(file):
            return file
        if isinstance(file, pathlib.Path):
            return file.with_suffix("")
        return os.path.splitext(file)[0]


class JinjaRendererAdapter(RendererAdapter):
    """Adapter for the Jinja2 template engine.

    Attributes:
        TEMPLATE_EXTENSION (str): The file extension used by Jinja2 templates.
    """

    TEMPLATE_EXTENSION = ".jinja"

    def __init__(self, template_dir: types.PathOrStr):
        """Initialize the JinjaRendererAdapter.
        Args:
            template_dir (PathOrStr): The directory where the templates are
                located.
        """
        self.set_template_dir(template_dir)

    def set_template_dir(self, template_dir: types.PathOrStr) -> None:
        """Set the directory where the templates are located.

        Sets the directory where the templates are located and initializes the
        Jinja2 environment and loader.

        Args:
            template_dir (PathOrStr): The directory where the templates are
                located.
        """
        self.template_dir = pathlib.Path(template_dir)

        loader = jinja2.FileSystemLoader(self.template_dir)
        self.environment = jinja2.Environment(loader=loader,
                                              undefined=jinja2.StrictUndefined)

    def render(self, template_name: types.PathOrStr,
               target_file: Union[types.PathOrStr, IO], **render_args):
        """Render a template to a file.

        The template with the given `template_name` is rendered to a file named
        `dest_name`, using the given render arguments. If the template is not
        found, a jinja2.exceptions.TemplateNotFound exception will be raised.
        If the file pointed by `template_name` is not a template, the file will
        be copied to `dest_name` as is.

        Args:
            template_name (str): The name of the template file to be rendered.
            dest_name (str or Path): The name of the destination file.
            render_args: Keyword arguments that will be passed to the template
                renderer.
        """
        # jinja uses posix paths (forward slashes) to find templates,
        # even on Windows
        template_name = pathlib.Path(template_name).as_posix()
        template = self.environment.get_template(template_name)
        stream = template.stream(**render_args)

        if isinstance(target_file, pathlib.Path):
            stream.dump(str(target_file))
        else:
            stream.dump(target_file)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class ExecuterTrackerAPIConnectionInfo(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Information sent to the executer tracker after registration.
    """

    class MetaOapg:
        required = {
            "redis_consumer_group",
            "redis_consumer_name",
            "redis_streams",
            "uuid",
        }

        class properties:
            uuid = schemas.StrSchema

            class redis_streams(schemas.ListSchema):

                class MetaOapg:
                    items = schemas.StrSchema

                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[
                        MetaOapg.items,
                        str,
                    ]], typing.List[typing.Union[
                        MetaOapg.items,
                        str,
                    ]]],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                ) -> 'redis_streams':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )

                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)

            redis_consumer_name = schemas.StrSchema
            redis_consumer_group = schemas.StrSchema
            __annotations__ = {
                "uuid": uuid,
                "redis_streams": redis_streams,
                "redis_consumer_name": redis_consumer_name,
                "redis_consumer_group": redis_consumer_group,
            }

    redis_consumer_group: MetaOapg.properties.redis_consumer_group
    redis_consumer_name: MetaOapg.properties.redis_consumer_name
    redis_streams: MetaOapg.properties.redis_streams
    uuid: MetaOapg.properties.uuid

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["uuid"]
    ) -> MetaOapg.properties.uuid:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["redis_streams"]
    ) -> MetaOapg.properties.redis_streams:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["redis_consumer_name"]
    ) -> MetaOapg.properties.redis_consumer_name:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["redis_consumer_group"]
    ) -> MetaOapg.properties.redis_consumer_group:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "uuid",
        "redis_streams",
        "redis_consumer_name",
        "redis_consumer_group",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["uuid"]
    ) -> MetaOapg.properties.uuid:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["redis_streams"]
    ) -> MetaOapg.properties.redis_streams:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["redis_consumer_name"]
    ) -> MetaOapg.properties.redis_consumer_name:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["redis_consumer_group"]
    ) -> MetaOapg.properties.redis_consumer_group:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "uuid",
        "redis_streams",
        "redis_consumer_name",
        "redis_consumer_group",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        redis_consumer_group: typing.Union[
            MetaOapg.properties.redis_consumer_group,
            str,
        ],
        redis_consumer_name: typing.Union[
            MetaOapg.properties.redis_consumer_name,
            str,
        ],
        redis_streams: typing.Union[
            MetaOapg.properties.redis_streams,
            list,
            tuple,
        ],
        uuid: typing.Union[
            MetaOapg.properties.uuid,
            str,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'ExecuterTrackerAPIConnectionInfo':
        return super().__new__(
            cls,
            *_args,
            redis_consumer_group=redis_consumer_group,
            redis_consumer_name=redis_consumer_name,
            redis_streams=redis_streams,
            uuid=uuid,
            _configuration=_configuration,
            **kwargs,
        )



"""CLI commands for listing both the active and available machine groups."""

from typing import TextIO
import argparse
import sys

from inductiva import resources, _cli
from inductiva.resources.machine_types import ProviderType


def pretty_print_machines_info(machines_dict):
    """Format and print the information for given machine dict."""

    print()

    for family, memory_types in machines_dict.items():
        print(f"CPU family: {family}")
        for memory, details in memory_types.items():
            vcpus = sorted([int(k) for k in details["vcpus"]], key=int)
            lssd = "(-lssd)" if details["lssd"] else ""
            print(f"   > {family}-{memory}- {(vcpus)} {lssd}")
        print()


def list_machine_types_available(args):
    """List available machine types information per provider and CPU series."""

    provider = args.provider
    machine_family = args.family

    machines_dict = {}

    resources_available = resources.machine_types.get_available_machine_types(
        provider, machine_family)

    for machine in resources_available:
        # Avoid the duplication of adding machine types with spot instances.
        if machine["spot"]:
            continue

        machine_type = machine["machine_type"]
        machine_info = machine_type.split("-")
        machine_family = machine_info[0]
        memory_type = machine_info[1]
        vcpus = machine_info[2] if len(machine_info) > 2 else None
        lssd = True if len(machine_info) > 3 else None

        if machine_family not in machines_dict:
            machines_dict[machine_family] = {}

        if memory_type not in machines_dict[machine_family]:
            machines_dict[machine_family][memory_type] = {
                "vcpus": [],
                "lssd": False
            }

        if vcpus is not None and vcpus not in machines_dict[machine_family][
                memory_type]["vcpus"]:
            machines_dict[machine_family][memory_type]["vcpus"].append(vcpus)

        if lssd is not None:
            machines_dict[machine_family][memory_type]["lssd"] = True

    pretty_print_machines_info(machines_dict)


def list_machine_groups(unused_args, fout: TextIO = sys.stdout):
    """List Resources."""
    resources.machine_groups.list(fout=fout)


def register(parser):
    """Register the list resources commands."""

    subparser = parser.add_parser("available",
                                  help="List available machine types.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.add_argument("-p",
                           "--provider",
                           type=str,
                           default="gcp",
                           choices=[t.value for t in ProviderType],
                           help="Filter the available types by provider.")
    subparser.add_argument("-f",
                           "--family",
                           default=None,
                           type=str,
                           help="Filter the available types by CPU series.")

    subparser.description = (
        "The `inductiva available` command provides a utility for listing the\n"
        "available machine types by provider (default: gcp) and CPU series.\n"
        "The list includes a description of the memory types and vCPUs "
        "available.\n\n")

    subparser.set_defaults(func=list_machine_types_available)

    subparser = parser.add_parser("list",
                                  aliases=["ls"],
                                  help="List currently active resources.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    _cli.utils.add_watch_argument(subparser)

    subparser.description = (
        "The `inductiva resources list` command provides a snapshot "
        "of your active computational resources.\nFrom machine type "
        "to start time, it gives you a comprehensive overview of "
        "your resources in one place.")
    subparser.set_defaults(func=list_machine_groups)



"""Custom logging functions"""
import os

import logging.handlers
import logging
import sys

import pathlib
import platform

from inductiva import constants

root_logger = logging.getLogger()


def get_logs_file_path():
    system = platform.system()
    logs_name = "inductiva.log"
    if system.lower() == "windows":
        logs_file_path = pathlib.Path.home(
        ) / "AppData" / "inductiva" / logs_name
    elif system.lower() in ["linux", "darwin"]:
        logs_file_path = constants.HOME_DIR / logs_name
    else:
        raise RuntimeError(f"Current operating system {system} not supported.")
    return logs_file_path


class NoExceptionFormatter(logging.Formatter):

    def format(self, record):
        record.exc_text = record.exc_info = None
        return super().format(record)


def log_and_exit(logger, level, msg, *args, **kwargs):
    logger.log(level, msg, *args, **kwargs)
    sys.exit(1)


def handle_uncaught_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    root_logger.error(
        "System encountered the following unhandled exception:\n"
        "%s\n  Exiting with code 1.",
        exc_value,
        exc_info=(exc_type, exc_value, exc_traceback))


def setup(level=logging.INFO):
    formatters = [
        logging.Formatter(
            fmt=
            "%(asctime)s|%(levelname)s|%(name)s|%(filename)s:%(lineno)"\
            "d %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S"),
        NoExceptionFormatter(fmt="%(message)s")
    ]

    logs_file_path = get_logs_file_path()
    logs_file_dir = logs_file_path.parent
    if not os.path.exists(logs_file_dir):
        os.mkdir(logs_file_dir)

    handlers = [
        logging.handlers.RotatingFileHandler(logs_file_path,
                                             encoding="utf8",
                                             maxBytes=1e6,
                                             backupCount=10),
        logging.StreamHandler()
    ]

    for handler, formatter in zip(handlers, formatters):
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)

    root_logger.setLevel(level)
    sys.excepthook = handle_uncaught_exception



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class User(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "bucket_name",
            "expiry_ts",
            "email",
            "username",
        }
        
        class properties:
            email = schemas.StrSchema
            username = schemas.StrSchema
            bucket_name = schemas.StrSchema
            expiry_ts = schemas.DateTimeSchema
            is_admin = schemas.BoolSchema
            __annotations__ = {
                "email": email,
                "username": username,
                "bucket_name": bucket_name,
                "expiry_ts": expiry_ts,
                "is_admin": is_admin,
            }
    
    bucket_name: MetaOapg.properties.bucket_name
    expiry_ts: MetaOapg.properties.expiry_ts
    email: MetaOapg.properties.email
    username: MetaOapg.properties.username
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["email"]) -> MetaOapg.properties.email: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["username"]) -> MetaOapg.properties.username: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["bucket_name"]) -> MetaOapg.properties.bucket_name: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["expiry_ts"]) -> MetaOapg.properties.expiry_ts: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["is_admin"]) -> MetaOapg.properties.is_admin: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["email", "username", "bucket_name", "expiry_ts", "is_admin", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["email"]) -> MetaOapg.properties.email: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["username"]) -> MetaOapg.properties.username: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["bucket_name"]) -> MetaOapg.properties.bucket_name: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["expiry_ts"]) -> MetaOapg.properties.expiry_ts: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["is_admin"]) -> typing.Union[MetaOapg.properties.is_admin, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["email", "username", "bucket_name", "expiry_ts", "is_admin", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        bucket_name: typing.Union[MetaOapg.properties.bucket_name, str, ],
        expiry_ts: typing.Union[MetaOapg.properties.expiry_ts, str, datetime, ],
        email: typing.Union[MetaOapg.properties.email, str, ],
        username: typing.Union[MetaOapg.properties.username, str, ],
        is_admin: typing.Union[MetaOapg.properties.is_admin, bool, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'User':
        return super().__new__(
            cls,
            *_args,
            bucket_name=bucket_name,
            expiry_ts=expiry_ts,
            email=email,
            username=username,
            is_admin=is_admin,
            _configuration=_configuration,
            **kwargs,
        )



"""Utility class to manage files.

Provides utility classes and functions for managing files and directories
and rendering template files for generation of simulation inputs.
"""
from typing import Iterable, Optional
import logging
import pathlib
import shutil
import glob
import re
import os

from inductiva import types
from inductiva.utils import format_utils

SUFFIX_SEPARATOR = "__"

_logger = logging.getLogger(__name__)


class FileManager:
    """Class for file management."""

    DEFAULT_ROOT_DIR = "input_dir"
    __root_dir = None  # pylint: disable=invalid-name

    def __check_root_dir(self) -> pathlib.Path:
        """Set default root folder if unset."""
        if self.__root_dir is None:
            self.set_root_dir(self.DEFAULT_ROOT_DIR)
        return self.__root_dir

    def set_root_dir(self, root_dir: types.PathOrStr):
        """Set a root directory for the file manager.

        All files managed through the manager will be
        placed inside a newly created folder with the given
        name. If a folder with the same name already exists,
        the newly created one will be appended with the
        "__1" suffix. If folders having the "__N" suffix already
        exist, the given name will be appended with "__{N+1}"
        where N is determined by the directory with the largest N.

        Args:
            root_dir (types.PathOrStr): Path to the root directory.
                If None, an error is raised.
        """

        if root_dir is None:
            raise ValueError("Given root directory cannot be None")
        root_dir = pathlib.Path(root_dir).resolve()

        if root_dir.is_dir():
            if format_utils.getenv_bool(
                    "INDUCTIVA_DISABLE_FILEMANAGER_AUTOSUFFIX", False):
                raise FileExistsError(f"Directory {root_dir} already exists.")
            generated_root_dir = _gen_unique_name(str(root_dir))
            _logger.info(
                "Directory %s already exists."
                " Setting root folder to %s.", root_dir, generated_root_dir)
            root_dir = pathlib.Path(generated_root_dir).resolve()

        _logger.info("Setting root folder to %s.", root_dir)
        root_dir.mkdir(parents=True, exist_ok=True)
        self.__root_dir = root_dir

    def get_root_dir(self):
        """Get the active root directory for the file manager."""

        return self.__check_root_dir()

    def copy_file(self,
                  source_file: types.PathOrStr,
                  target_file: Optional[types.PathOrStr] = None,
                  overwrite: bool = False) -> pathlib.Path:
        """Copy a file to the root_dir.

        This method copies the contents of the `source_file` to a `target_file`
        inside the root directory. If `target_file` is None, the copied file is
        named as the source_file, but not the parent directory structure.

        In case, the file with the same name provided already exists, the
        method by default raises a FileExistsError. To overwrite it anyway, set
        the flag `overwrite` to True.

        Args:
            source_file (types.PathOrStr): Path to the source file.
            target_file (types.PathOrStr - Optional): Path to the target file.
            overwrite (bool): Overwrite the target file if it already exists.
        """

        root_dir = self.__check_root_dir()

        if target_file is None:
            target_file = pathlib.Path(source_file).name
        target_file = root_dir / target_file

        if not overwrite and target_file.exists():
            raise FileExistsError(f"{target_file} already exists.")

        target_file.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(source_file, target_file)

        return target_file

    def copy_dir(self,
                 source_dir: types.PathOrStr,
                 target_dir: Optional[types.PathOrStr] = None,
                 overwrite: bool = False) -> types.PathOrStr:
        """Copy a directory to the root_dir.

        This method copies the contents of the `source_dir` to a `target_dir`
        inside the root directory, keeping the structure as the source
        directory. If `target_dir` is None, the source directory is copied to
        the root directory.

        If any file in the target directory has the same name as the ones
        about to be copied, the default behavior of this method is to raise a
        FileExistsError. To overwrite the target directory anyway, set the flag
        `overwrite` to True.

        Args:
            source_dir (types.PathOrStr): Path to the source directory.
            target_dir (types.PathOrStr - Optional): Path to the target
                directory.
            overwrite (bool): Overwrite the target directory if  already exists.
        """
        root_dir = self.__check_root_dir()

        if target_dir is None:
            target_dir = "."

        target_dir = root_dir / target_dir
        if not overwrite:
            try:
                self._check_precopy_dir(source_dir, target_dir)
            except FileExistsError as e:
                msg = f"{e}; set `overwrite=True` to overwrite existing files."
                raise FileExistsError(msg) from e

        shutil.copytree(source_dir,
                        target_dir,
                        dirs_exist_ok=True,
                        symlinks=True)

        return target_dir

    def _check_precopy_dir(self, source_dir: pathlib.Path,
                           target_dir: pathlib.Path):
        """Check if any file in source_dir exists in target_dir.

        Check if the target filenames exist and raise a FileExistsError if
        any of them already exists.

        Args:
            source_dir (pathlib.Path): The source directory.
            target_dir (pathlib.Path): The target directory.
        """
        target_dir = pathlib.Path(target_dir)

        for dirpath, _, files in os.walk(source_dir, topdown=True):
            target_subdir = target_dir / os.path.relpath(dirpath, source_dir)
            for file in files:
                if (target_subdir / file).exists():
                    raise FileExistsError(
                        f"File {file} already exists in {target_subdir}")


def _gen_unique_suffix(name: str, filenames: Iterable[str]) -> str:
    """Generate a suffix for a filename, based on a list of files.

    Args:
        name: Name of the file.
        filenames: List of filenames.
    """
    regex = f"^{name}{SUFFIX_SEPARATOR}([0-9]+$)"

    max_suffix = -1
    for filename in filenames:
        if match := re.match(regex, filename):
            max_suffix = max(max_suffix, int(match.group(1)))

    exists_solo = name in filenames
    exists_derived = max_suffix >= 0

    if exists_derived:
        suffix = f"{SUFFIX_SEPARATOR}{max_suffix+1}"
    elif exists_solo:
        suffix = f"{SUFFIX_SEPARATOR}2"
    else:
        suffix = ""
    return suffix


def _gen_unique_name(name: str) -> str:
    """Generates an unique folder name.

    Generates an unique folder name derived from the given name and
    based on the names of the existing folders in the current working directory.
    If no folder exists with the same name as the given name,
    the input name will be returned; otherwise, a suffix is appended
    to the input name to make it unique.

    See `_gen_unique_suffix` for more details.
    """

    name = name.strip()
    filenames = glob.glob(f"{name}*")
    return name + _gen_unique_suffix(name, filenames)



from inductiva.client.paths.storage_contents.get import ApiForget
from inductiva.client.paths.storage_contents.delete import ApiFordelete


class StorageContents(
        ApiForget,
        ApiFordelete,
):
    pass



# coding: utf-8

# flake8: noqa

# import all models into this package
# if you have many models here with many references from one model to another this may
# raise a RecursionError
# to avoid this, import only the models that you directly need like:
# from inductiva.client.model.pet import Pet
# or import this package, but before doing it, use:
# import sys
# sys.setrecursionlimit(n)

from inductiva.client.model.backend_version import BackendVersion
from inductiva.client.model.base_machine_type import BaseMachineType
from inductiva.client.model.base_vm_group import BaseVMGroup
from inductiva.client.model.body_upload_task_input import BodyUploadTaskInput
from inductiva.client.model.default_machine_group import DefaultMachineGroup
from inductiva.client.model.executer import Executer
from inductiva.client.model.executer_tracker_api_connection_info import ExecuterTrackerAPIConnectionInfo
from inductiva.client.model.executer_tracker_register_info import ExecuterTrackerRegisterInfo
from inductiva.client.model.file_info import FileInfo
from inductiva.client.model.http_validation_error import HTTPValidationError
from inductiva.client.model.machine_type import MachineType
from inductiva.client.model.output_archive_info import OutputArchiveInfo
from inductiva.client.model.provider import Provider
from inductiva.client.model.providers import Providers
from inductiva.client.model.quota import Quota
from inductiva.client.model.task import Task
from inductiva.client.model.task_request import TaskRequest
from inductiva.client.model.task_status import TaskStatus
from inductiva.client.model.task_status_code import TaskStatusCode
from inductiva.client.model.user import User
from inductiva.client.model.user_api_key import UserApiKey
from inductiva.client.model.user_create import UserCreate
from inductiva.client.model.vm_group_config import VMGroupConfig
from inductiva.client.model.vm_group_lifecycle_config import VMGroupLifecycleConfig
from inductiva.client.model.validation_error import ValidationError
from inductiva.client.model.version_comparison_result import VersionComparisonResult



"""CLI for logs."""
import sys

from .. import utils as cli_utils
from ... import tasks


def stream_task_logs(args):
    """Consume the stream logs of a certain task."""
    task_id = args.task_id
    task = tasks.Task(task_id)

    if not task.is_running():
        print(
            f"The current status of task {task_id} is '{task.get_status()}'\n"
            "and the simulation logs are not available for streaming.\n"
            "For more information about the task status, use:\n\n"
            f"  inductiva tasks list --task-id {task_id}\n",
            file=sys.stderr)
        return 1

    consumer = tasks.streams.TaskStreamConsumer(task_id)
    consumer.run_forever()
    return 0


def register(parser):
    cli_utils.show_help_msg(parser)

    parser.add_argument("task_id",
                        type=str,
                        help="ID of the task for which to consume the stream.")

    # Register function to call when this subcommand is used
    parser.set_defaults(func=stream_task_logs)



"""Functionality to interact with the Web API."""
from .methods import download_output, invoke_async_api, get_task_status
from .methods import validate_api_key
from .methods import get_client
from .methods import compare_client_and_backend_versions



import typing_extensions

from inductiva.client.paths import PathValues
from inductiva.client.apis.paths.tasks_auth import TasksAuth
from inductiva.client.apis.paths.tasks_submit import TasksSubmit
from inductiva.client.apis.paths.tasks_task_id_input import TasksTaskIdInput
from inductiva.client.apis.paths.tasks_task_id import TasksTaskId
from inductiva.client.apis.paths.tasks import Tasks
from inductiva.client.apis.paths.tasks_task_id_status import TasksTaskIdStatus
from inductiva.client.apis.paths.tasks_task_id_output_list import TasksTaskIdOutputList
from inductiva.client.apis.paths.tasks_task_id_output import TasksTaskIdOutput
from inductiva.client.apis.paths.tasks_task_id_kill import TasksTaskIdKill
from inductiva.client.apis.paths.admin_users import AdminUsers
from inductiva.client.apis.paths.admin_users_email_api_key import AdminUsersEmailApiKey
from inductiva.client.apis.paths.admin_users_email import AdminUsersEmail
from inductiva.client.apis.paths.admin_users_email_expiry_ts import AdminUsersEmailExpiryTs
from inductiva.client.apis.paths.admin_users_username_tasks import AdminUsersUsernameTasks
from inductiva.client.apis.paths.admin_groups import AdminGroups
from inductiva.client.apis.paths.admin_groups_active import AdminGroupsActive
from inductiva.client.apis.paths.admin_groups_default import AdminGroupsDefault
from inductiva.client.apis.paths.admin_groups_default_machine_group_id import AdminGroupsDefaultMachineGroupId
from inductiva.client.apis.paths.admin_providers import AdminProviders
from inductiva.client.apis.paths.admin_providers_provider_id import AdminProvidersProviderId
from inductiva.client.apis.paths.admin_active_tasks import AdminActiveTasks
from inductiva.client.apis.paths.executer_tracker_register import ExecuterTrackerRegister
from inductiva.client.apis.paths.compute_group import ComputeGroup
from inductiva.client.apis.paths.compute_type import ComputeType
from inductiva.client.apis.paths.compute_group_start import ComputeGroupStart
from inductiva.client.apis.paths.compute_group_mg_id import ComputeGroupMgId
from inductiva.client.apis.paths.compute_price import ComputePrice
from inductiva.client.apis.paths.compute_groups import ComputeGroups
from inductiva.client.apis.paths.compute_group_status import ComputeGroupStatus
from inductiva.client.apis.paths.compute_machine_types import ComputeMachineTypes
from inductiva.client.apis.paths.compute_group_machine import ComputeGroupMachine
from inductiva.client.apis.paths.compute_group_name import ComputeGroupName
from inductiva.client.apis.paths.storage_size import StorageSize
from inductiva.client.apis.paths.storage_contents import StorageContents
from inductiva.client.apis.paths.version import Version
from inductiva.client.apis.paths.version_check import VersionCheck
from inductiva.client.apis.paths.users_quotas import UsersQuotas

PathToApi = typing_extensions.TypedDict(
    'PathToApi', {
        PathValues.TASKS_AUTH:
            TasksAuth,
        PathValues.TASKS_SUBMIT:
            TasksSubmit,
        PathValues.TASKS_TASK_ID_INPUT:
            TasksTaskIdInput,
        PathValues.TASKS_TASK_ID:
            TasksTaskId,
        PathValues.TASKS:
            Tasks,
        PathValues.TASKS_TASK_ID_STATUS:
            TasksTaskIdStatus,
        PathValues.TASKS_TASK_ID_OUTPUT_LIST:
            TasksTaskIdOutputList,
        PathValues.TASKS_TASK_ID_OUTPUT:
            TasksTaskIdOutput,
        PathValues.TASKS_TASK_ID_KILL:
            TasksTaskIdKill,
        PathValues.ADMIN_USERS:
            AdminUsers,
        PathValues.ADMIN_USERS_EMAIL_API_KEY:
            AdminUsersEmailApiKey,
        PathValues.ADMIN_USERS_EMAIL:
            AdminUsersEmail,
        PathValues.ADMIN_USERS_EMAIL_EXPIRY_TS:
            AdminUsersEmailExpiryTs,
        PathValues.ADMIN_USERS_USERNAME_TASKS:
            AdminUsersUsernameTasks,
        PathValues.ADMIN_GROUPS:
            AdminGroups,
        PathValues.ADMIN_GROUPS_ACTIVE:
            AdminGroupsActive,
        PathValues.ADMIN_GROUPS_DEFAULT:
            AdminGroupsDefault,
        PathValues.ADMIN_GROUPS_DEFAULT_MACHINE_GROUP_ID:
            AdminGroupsDefaultMachineGroupId,
        PathValues.ADMIN_PROVIDERS:
            AdminProviders,
        PathValues.ADMIN_PROVIDERS_PROVIDER_ID:
            AdminProvidersProviderId,
        PathValues.ADMIN_ACTIVE_TASKS:
            AdminActiveTasks,
        PathValues.EXECUTERTRACKER_REGISTER:
            ExecuterTrackerRegister,
        PathValues.COMPUTE_GROUP:
            ComputeGroup,
        PathValues.COMPUTE_TYPE:
            ComputeType,
        PathValues.COMPUTE_GROUP_START:
            ComputeGroupStart,
        PathValues.COMPUTE_GROUP_MG_ID:
            ComputeGroupMgId,
        PathValues.COMPUTE_PRICE:
            ComputePrice,
        PathValues.COMPUTE_GROUPS:
            ComputeGroups,
        PathValues.COMPUTE_GROUP_STATUS:
            ComputeGroupStatus,
        PathValues.COMPUTE_MACHINE_TYPES:
            ComputeMachineTypes,
        PathValues.COMPUTE_GROUP_MACHINE:
            ComputeGroupMachine,
        PathValues.COMPUTE_GROUP_NAME:
            ComputeGroupName,
        PathValues.STORAGE_SIZE:
            StorageSize,
        PathValues.STORAGE_CONTENTS:
            StorageContents,
        PathValues.VERSION:
            Version,
        PathValues.VERSIONCHECK:
            VersionCheck,
        PathValues.USERS_QUOTAS:
            UsersQuotas,
    })

path_to_api = PathToApi({
    PathValues.TASKS_AUTH:
        TasksAuth,
    PathValues.TASKS_SUBMIT:
        TasksSubmit,
    PathValues.TASKS_TASK_ID_INPUT:
        TasksTaskIdInput,
    PathValues.TASKS_TASK_ID:
        TasksTaskId,
    PathValues.TASKS:
        Tasks,
    PathValues.TASKS_TASK_ID_STATUS:
        TasksTaskIdStatus,
    PathValues.TASKS_TASK_ID_OUTPUT_LIST:
        TasksTaskIdOutputList,
    PathValues.TASKS_TASK_ID_OUTPUT:
        TasksTaskIdOutput,
    PathValues.TASKS_TASK_ID_KILL:
        TasksTaskIdKill,
    PathValues.ADMIN_USERS:
        AdminUsers,
    PathValues.ADMIN_USERS_EMAIL_API_KEY:
        AdminUsersEmailApiKey,
    PathValues.ADMIN_USERS_EMAIL:
        AdminUsersEmail,
    PathValues.ADMIN_USERS_EMAIL_EXPIRY_TS:
        AdminUsersEmailExpiryTs,
    PathValues.ADMIN_USERS_USERNAME_TASKS:
        AdminUsersUsernameTasks,
    PathValues.ADMIN_GROUPS:
        AdminGroups,
    PathValues.ADMIN_GROUPS_ACTIVE:
        AdminGroupsActive,
    PathValues.ADMIN_GROUPS_DEFAULT:
        AdminGroupsDefault,
    PathValues.ADMIN_GROUPS_DEFAULT_MACHINE_GROUP_ID:
        AdminGroupsDefaultMachineGroupId,
    PathValues.ADMIN_PROVIDERS:
        AdminProviders,
    PathValues.ADMIN_PROVIDERS_PROVIDER_ID:
        AdminProvidersProviderId,
    PathValues.ADMIN_ACTIVE_TASKS:
        AdminActiveTasks,
    PathValues.EXECUTERTRACKER_REGISTER:
        ExecuterTrackerRegister,
    PathValues.COMPUTE_GROUP:
        ComputeGroup,
    PathValues.COMPUTE_TYPE:
        ComputeType,
    PathValues.COMPUTE_GROUP_START:
        ComputeGroupStart,
    PathValues.COMPUTE_GROUP_MG_ID:
        ComputeGroupMgId,
    PathValues.COMPUTE_PRICE:
        ComputePrice,
    PathValues.COMPUTE_GROUPS:
        ComputeGroups,
    PathValues.COMPUTE_GROUP_STATUS:
        ComputeGroupStatus,
    PathValues.COMPUTE_MACHINE_TYPES:
        ComputeMachineTypes,
    PathValues.COMPUTE_GROUP_MACHINE:
        ComputeGroupMachine,
    PathValues.COMPUTE_GROUP_NAME:
        ComputeGroupName,
    PathValues.STORAGE_SIZE:
        StorageSize,
    PathValues.STORAGE_CONTENTS:
        StorageContents,
    PathValues.VERSION:
        Version,
    PathValues.VERSIONCHECK:
        VersionCheck,
    PathValues.USERS_QUOTAS:
        UsersQuotas,
})



"""Inductiva template manager class to manipulate files and render templates."""

from typing import Optional, Type
import pathlib
import shutil
import os

from inductiva import types
from inductiva.utils import file_manager
from inductiva.templating import renderers


class TemplateManager(file_manager.FileManager):
    """Template manager for rendering files and directories.

    This class extends functionality of the FileManager to render templates
    files. It provides methods to render single files and entire directories,
    """

    def __init__(self,
                 template_dir: types.PathOrStr,
                 root_dir: types.PathOrStr = "rendered_dir",
                 renderer_cls: Type[renderers.RendererAdapter] = None):
        """Initialize the template manager object.

        Args:
            template_dir (types.PathOrStr): Path to the template directory.
            root_dir (types.PathOrStr): Path to the default directory where
                rendered files will be written to.
            renderer (Type[renderers.RendererAdapter]): A Template Renderer
                class to use for rendering the templates. If not provided, the
                default renderer is used.
        """
        if not root_dir:
            raise ValueError("Root directory must be valid path or string")

        renderer_cls = renderer_cls or renderers.JinjaRendererAdapter
        self.set_root_dir(root_dir)
        self.renderer = renderer_cls(template_dir)
        self.template_dir = pathlib.Path(template_dir)
        self.template_dir_struct = self._gen_dir_struture()

    def _gen_dir_struture(self):
        """
        Generate a dictionary with the structure of the template directory.
        The keys are the subdirectories of the template directory, relative to
        the template directory itself. The values are dictionaries with the keys
        "files" and "templates" containing the lists of the files and templates
        in each subdirectory, respectively.
        Example:
        >>> self._gen_dir_struture()
        {
            ".": {
                "files": ["file0.1"],
                "templates": ["template0.1"]
            },
            "subdir1": {
                "files": ["file1.1", "file1.2"],
                "templates": ["template1.1", "template1.2"]
            }
            ...
        }
        """
        is_template = self.renderer.is_template
        return {
            os.path.relpath(root, start=self.template_dir): {
                "templates": [f for f in files if is_template(f)],
                "files": [f for f in files if not is_template(f)],
            } for root, _, files in os.walk(self.template_dir, topdown=True)
        }

    def render_file(self,
                    source_file: types.PathOrStr,
                    target_file: Optional[types.PathOrStr] = None,
                    overwrite: bool = False,
                    **render_args) -> pathlib.Path:
        """Render a single template to a file.

        Renders the `source_file` from the template directory to `target_file`
        inside the root directory, using the given render arguments.
        If `target_file` is not given, the name of `source_file` is used and the
        file is rendered directly to the root directory. When `overwrite` is
        False (default) and the destination files already exists, a
        FileExistsError will be raised.

        Args:
            source_file (str): The relative path of the template file inside
                the template directory.
            target_file: The name of the destination file. If not given, the
                name of the template will be used and the rendered file will be
                dumped to the root directory.
            overwrite: If True, the destination file will be overwritten if it
                already exists. If False (default), a FileExistsError will be
                raised if the destination file already exists.
            render_args: Keyword arguments that will be passed to the template
                renderer.

        Raises:
            FileExistsError: If the destination file already exists and
                `overwrite` is False.
        """

        if not self.renderer.is_template(source_file):
            raise ValueError(f"{source_file} is not a template file.")

        if target_file is None:
            target_file = pathlib.Path(source_file).name
        target_file = pathlib.Path(target_file)

        target_file = self.renderer.strip_extension(target_file)
        target_file = self.get_root_dir() / target_file

        if not overwrite and target_file.exists():
            raise FileExistsError(f"Target file {target_file} already exists.")

        target_file.parent.mkdir(parents=True, exist_ok=True)
        self.renderer.render(template_name=source_file,
                             target_file=target_file,
                             **render_args)

        return target_file

    def render_dir(self,
                   source_dir: Optional[types.PathOrStr] = None,
                   target_dir: Optional[types.PathOrStr] = None,
                   overwrite: bool = False,
                   **render_args) -> pathlib.Path:
        """
        Render an entire template directory.

        Renders the entire template directory to the given `target_dir`,
        relative to the root directory, using the provided rendering args.
        If `source_dir` is not provided (default), the entire template directory
        is rendered. If `target_dir` is provided, is is assumed it is relative
        to the template directory, in which case only that sub-directory is
        rendered. I any case, the tree structure of the `source_dir` is
        preserved in the `target_dir`.

        When `target_dir` does not exist, it will be created. If any destination
        file inside `target_dir` already exists, a FileExistsError will be
        raised when `overwrite` is False (default). Otherwise, files are
        overwritten.

        Args:
            source_dir (types.PathOrStr): Path to the source directory, relative
                to the template directory. If None, the full template directory
                is rendered.
            target_dir (types.PathOrStr): Path to the target directory, relative
                to the root directory. If None, the root directory is used.
                If set to an absolute path, the root directory is not used.
            overwrite (bool): If True, the destination files will be overwritten
                if they already exist. If False (default), a FileExistsError
                will be raised if any destination file already exists.
            render_args: Keyword arguments to render the template files.

        Raises:
            FileExistsError: If the destination file already exists and
                `overwrite` is False.
        """

        if source_dir is None:
            source_dir = "."
        source_dir = pathlib.Path(source_dir)

        if target_dir is None:
            target_dir = "."
        target_dir = pathlib.Path(target_dir)

        target_dir = self.get_root_dir() / target_dir

        if not overwrite:
            try:
                self._check_prerender_dir(target_dir)
            except FileExistsError as e:
                msg = f"{e}; set `overwrite=True` to overwrite existing files."
                raise FileExistsError(msg) from e

        for subdir, contents in self.template_dir_struct.items():
            target_subdir = target_dir / subdir
            target_subdir.mkdir(parents=True, exist_ok=True)
            source_dir = self.template_dir / subdir

            # simply copy non-template files to destination
            for file in contents["files"]:
                shutil.copy(source_dir / file, target_subdir)

            # render template files
            for file in contents["templates"]:
                target = target_subdir / self.renderer.strip_extension(file)
                template_name = (pathlib.Path(subdir) / file).as_posix()
                self.renderer.render(template_name, target, **render_args)

        return target_dir

    def _check_prerender_dir(self, target_dir: pathlib.Path):
        """Check if the destination filenames exist.

        Check if the destination filenames exist and raise a FileExistsError if
        any of them already exists. For template files, the extension is
        stripped before checking.

        Args:
            target_dir (pathlib.Path): The destination directory.
        """
        for subdir, contents in self.template_dir_struct.items():
            dest_subdir = target_dir / subdir
            for file in contents["files"] + contents["templates"]:
                target_name = self.renderer.strip_extension(dest_subdir / file)
                if target_name.exists():
                    raise FileExistsError(f"File {target_name} already exists.")



"""DualSPHysics simulator module of the API."""

from typing import Optional

from inductiva import types, tasks, simulators


class DualSPHysics(simulators.Simulator):
    """Class to invoke a generic DualSPHysics simulation on the API."""

    def __init__(self):
        super().__init__()
        self.api_method_name = "sph.dualsphysics.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        commands: types.Commands,
        on: Optional[types.ComputationalResources] = None,
        storage_dir: Optional[types.Path] = "",
        extra_metadata: Optional[dict] = None,
    ) -> tasks.Task:
        """Executes a DualSPHysics simulation.

        Args:
            input_dir: Directory with simulation input files.
            sim_config_filename: Simulation config file.
            on: The computational resource to launch the simulation on. If None
                the simulation is submitted to a machine in the default pool.
            storage_dir: Directory for storing results.

        Returns:
            tasks.Task: An object representing the simulation task.
        """
        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           extra_metadata=extra_metadata)
